import os
import cv2
import random
import numpy as np
import tensorflow as tf
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
from collections import deque
from gameAPI.agent import gameAgent



class ActorCritic:
	def __init__(self, config, **kwargs):
			self.config = config
			self.actor_net, self.critic_net = self._create_networks()

	def _create_networks(self):
			# 输入层
			x = tf.placeholder(tf.float32, [None, *self.config.frame_size, 3])

			# 共享网络层
			conv1 = tf.layers.conv2d(inputs=x, filters=32, kernel_size=[8, 8], strides=4, activation=tf.nn.relu)
			conv2 = tf.layers.conv2d(inputs=conv1, filters=64, kernel_size=[4, 4], strides=2, activation=tf.nn.relu)
			conv3 = tf.layers.conv2d(inputs=conv2, filters=64, kernel_size=[3, 3], strides=1, activation=tf.nn.relu)
			flattened = tf.layers.flatten(conv3)

			# 演员网络
			fc_actor = tf.layers.dense(flattened, 512, activation=tf.nn.relu)
			action_probs = tf.layers.dense(fc_actor, self.config.num_actions, activation=tf.nn.softmax)

			# 评论家网络
			fc_critic = tf.layers.dense(flattened, 512, activation=tf.nn.relu)
			state_value = tf.layers.dense(fc_critic, 1)

			return (x, action_probs), (x, state_value)

    
	def train(self, session):
			# 创建网络
			(actor_in, action_probs), (critic_in, state_value) = self._create_networks()

			# 设置占位符
			actions_placeholder = tf.placeholder(tf.int32, [None])
			rewards_placeholder = tf.placeholder(tf.float32, [None])
			advantages_placeholder = tf.placeholder(tf.float32, [None])

			# 评论家网络的损失函数
			critic_loss = tf.reduce_mean(tf.square(rewards_placeholder - state_value))

			# 演员网络的损失函数
			action_masks = tf.one_hot(actions_placeholder, self.config.num_actions)
			picked_action_probs = tf.reduce_sum(action_probs * action_masks, axis=1)
			actor_loss = -tf.reduce_mean(tf.log(picked_action_probs) * advantages_placeholder)

			# 优化器
			actor_optimizer = tf.train.AdamOptimizer(self.config.lr).minimize(actor_loss)
			critic_optimizer = tf.train.AdamOptimizer(self.config.lr).minimize(critic_loss)

			saver = tf.train.Saver()
			session.run(tf.global_variables_initializer())

			# game agent
			game_agent = gameAgent()

			# 训练循环
			for episode in range(self.config.num_episodes):
				frame_info = game_agent.nextFrame(None)
				state = frame_info[0]  # 提取 frame
				state = cv2.resize(state, (96, 96))  # 调整 state 大小
				episode_rewards = []
				episode_actions = []
				episode_states = []
				episode_values = []
				done = False
				while not done:
					# 获取动作概率和状态价值
					action_probabilities, state_val = session.run([action_probs, state_value], feed_dict={actor_in: [state], critic_in: [state]})
					action_index = np.random.choice(range(3), p=action_probabilities[0])
					action = [0, 0, 0]
					action[action_index] = 1

					frame_info = game_agent.nextFrame(action)
					next_state = frame_info[0]
					next_state = cv2.resize(next_state, (96, 96))
					reward = frame_info[2]
					done = frame_info[3]

					episode_states.append(state)
					episode_actions.append(action_index)
					episode_rewards.append(reward)
					episode_values.append(state_val[0][0])
					state = next_state 

				# 计算累积回报和优势
				cumulative_rewards = self._compute_cumulative_rewards(episode_rewards)
				advantages = cumulative_rewards - np.array(episode_values)

				# 更新网络
				session.run(actor_optimizer, feed_dict={
					actor_in: episode_states,
					actions_placeholder: episode_actions,
					advantages_placeholder: advantages
				})
				session.run(critic_optimizer, feed_dict={
					critic_in: episode_states,
					rewards_placeholder: cumulative_rewards
				})

				if episode % self.config.save_interval == 0:
					saver.save(session, os.path.join(self.config.model_dir, 'ActorCritic_Pong'), global_step=episode)

	def _compute_cumulative_rewards(self, rewards):
			# 计算累积回报
			cumulative_rewards = np.zeros_like(rewards, dtype=np.float32)
			running_add = 0
			for t in reversed(range(len(rewards))):
				running_add = running_add * self.config.gamma + rewards[t]
				cumulative_rewards[t] = running_add
			return cumulative_rewards
